
package cds;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Simple exercise to experiment with tasks and tread pools.
 * A single task is provided both as a Runnable and as a Callable<String>.
 * Each task just extracts a random number and prints it out.
 * 
 * @author alessio
 */


public class CDS_03 {
    //constants to set up the desired test   
    static final int NUMTASKS = 7;
    static final boolean SERIALEXEC = false; //either serial execution or via thread pooling
    static final int NUMTHREADS = 3;
    static final boolean RUNNABLETEST = true; //either runnableTest or callableTest
    
    //the executor and the array with tasks are kept in private fields
    private final ExecutorService myExecutor;
    private final MyTrivialTask[] myTasks;

    //the executor to be used is passed to the constructor
    public CDS_03() {
        this(Executors.newSingleThreadExecutor());
    }
    
    public CDS_03(ExecutorService exc) {
        myExecutor = exc;       
        // all the tasks are created and put in the array "myTasks"
        myTasks = new MyTrivialTask[NUMTASKS];
        for (int i=0; i<myTasks.length; myTasks[i++]=new MyTrivialTask());  
    }
    

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        CDS_03 myInst;
        
        if(SERIALEXEC)
            myInst = new CDS_03();
        else
            myInst = new CDS_03(Executors.newFixedThreadPool(NUMTHREADS));

        if(RUNNABLETEST)
            myInst.doRunnableTest();
        else
            myInst.doCallableTest();          
       
    }
    
    private void doRunnableTest(){
        for(Runnable t: myTasks)
            myExecutor.execute(t);
        
        myExecutor.shutdown();
    }
    
    private void doCallableTest(){
        List<Future<String>> futureResults=new ArrayList<>();
        
        for(Callable c: myTasks)
            futureResults.add(myExecutor.submit(c));
                
        for(Future<String> f: futureResults) {
            try {
                System.out.println(f.get());
            } catch (InterruptedException | ExecutionException ex) {}
        }
        
        myExecutor.shutdown();
    } 
}


// this class implements the body for a task (either runnable or callable)
class MyTrivialTask implements Runnable, Callable<String> {
    
    static final int TASKDELAY = 3000; //milliseconds
    static final int MAXNUM = 12;
    java.util.Random myPRNG = new java.util.Random(); //thread-safe???
    static final AtomicInteger taskCounter = new AtomicInteger(0);
    final int taskID = taskCounter.getAndIncrement();

    @Override
    public void run() {
        System.out.println(call());
    }
    
    @Override
    public String call() {
        
        wasteTime(); //idle for a while, to promote different interleavings
        
        int extractedNumber = myPRNG.nextInt(MAXNUM+1);
                
        return message(extractedNumber);
    }
    
    
    private String message(int num){
        
        int myThreadID = ThreadSeqID.get(); //the ID for the thread running this task is obtained
        
        String msg = new String(new char[taskID]).replace("\0", ".......");
        
        msg = msg+"Task " + Integer.toString(taskID) + 
                " run by Thread "+ Integer.toString(myThreadID) + 
                " - # extracted: " + Integer.toString(num);
        return msg;    
    }
    
    private void wasteTime(){
        
        try {
            Thread.sleep( myPRNG.nextInt(TASKDELAY) );
        } catch (InterruptedException ex) {
            Thread.currentThread().interrupt(); //set its own "interrupted" flag
        }
    }
    
}

