package cds;

import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.*;

/**
 * Exercise to experiment a tread starvation deadlock, in case the pool 
 * is not sufficiently large.
 * A shared var c is initialized to INITVALUE; each task, as it starts, decrements 
 * it and waits in case its value is positive. Subsequently, it notifies one 
 * waiting thread.
 * Deadlock is experienced whenever c has no chance to reach negative values.
 * 
 * @author alessio
 */

public class ThreadStarvationDeadlockTest {
    static final int NUMTASKS=4;
    static final int POOLSIZE=4;
    static final int INITVALUE=+4; //initial number for the shared counter;
                          //number of threads to start before executing any task
    private final ExecutorService myExecutor;
    private final IncClass[] myTasks;
    
    public ThreadStarvationDeadlockTest(ExecutorService exc) {
        myExecutor = exc;       
        // all the tasks are created
        myTasks = new IncClass[NUMTASKS];
        for (int i=0; i<myTasks.length; myTasks[i++]=new IncClass());  
    }
    
    private void runAll(){
        for(Runnable t: myTasks)
            myExecutor.execute(t);
        myExecutor.shutdown();
    }
    
    public static void main(String[] arg){
        ThreadStarvationDeadlockTest myInst=
            new ThreadStarvationDeadlockTest(Executors.newFixedThreadPool(POOLSIZE));
        
        myInst.runAll();
    }
    


    static class IncClass implements Runnable{
        //a static lock, condition, and atomic var are shared across all the instances of this class
        private static final Lock sharedLock = new ReentrantLock();
        private static final Condition cond = sharedLock.newCondition();
        private static final AtomicInteger c = new AtomicInteger(INITVALUE);

        public void IncClass(){
        }

        @Override
        public void run() {
            int taskID = INITVALUE - c.getAndDecrement();
            
            //start message
            String msg = new String(new char[taskID]).replace("\0", "  ");
            msg += "Task #"+Integer.toString(taskID)+": STARTED!";
            System.out.println(msg);
        
            sharedLock.lock();
            try {
                while(c.get()>0) {
                    cond.await();  //"sleeping", i.e. suspension on cond. queue
                }
                cond.signal();
            }
            catch (InterruptedException ie) {
                Thread.currentThread().interrupt();
            }
            finally {
                sharedLock.unlock();
                
                //end message
                msg = new String(new char[taskID]).replace("\0", "  ");
                msg += "Task #"+Integer.toString(taskID)+": DONE!";
                System.out.println(msg);
            }

        }
    }

}
