package cds;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.CountDownLatch;

/**
 * Exercise to experiment a tread starvation deadlock, in case the pool 
 * is not sufficiently large: tasks interact with one another because a 
 * CountDouwnLatch is used to make a group of them meet.
 * 
 * Deadlock is experienced whenever the number of meeting tasks exceeds 
 * the pool size.
 * 
 * @author alessio
 */

public class ThreadStarvationDeadlockTestLatch {
    static final int NUMTASKS  =  4;
    static final int POOLSIZE  =  3;
    static final int GROUPSIZE =  3; //initial number for the latch
    
    private final ExecutorService myExecutor;
    private final IncClass[] myTasks;
    
    public ThreadStarvationDeadlockTestLatch(ExecutorService exc) {
        myExecutor = exc;       
        // all the tasks are created
        myTasks = new IncClass[NUMTASKS];
        for (int i=0; i<myTasks.length; myTasks[i++]=new IncClass());  
    }
    
    private void runAll(){
        for(Runnable t: myTasks)
            myExecutor.execute(t);
        myExecutor.shutdown();
    }
    
    public static void main(String[] arg){
        ThreadStarvationDeadlockTestLatch myInst =
            new ThreadStarvationDeadlockTestLatch(Executors.newFixedThreadPool(POOLSIZE));
        
        myInst.runAll();
    }
    


    static class IncClass implements Runnable{
        //a static lock, condition, and atomic var are shared across all the instances of this class
        private static final CountDownLatch sharedLatch = new CountDownLatch(GROUPSIZE);
        // var c is used only to give tasks an ID
        private static final AtomicInteger c = new AtomicInteger();
        private final int taskID = c.getAndIncrement();
        

        public void IncClass(){
        }

        @Override
        public void run() {
            
            //start message
            String msg = new String(new char[taskID]).replace("\0", "  ");
            msg += "Task #"+Integer.toString(taskID)+": STARTED!";
            System.out.println(msg);
        
            sharedLatch.countDown();
            try {
                sharedLatch.await();
            }
            catch (InterruptedException ie) {
                Thread.currentThread().interrupt();
            }
            finally {
                //end message
                msg = new String(new char[taskID]).replace("\0", "  ");
                msg += "Task #"+Integer.toString(taskID)+": DONE!";
                System.out.println(msg);
            }

        }
    }

}
